#!/usr/bin/env python3

from pydoc import locate

from agora.abc import ProcessABC


class PostProcessABC(ProcessABC):
    """
    Extend ProcessABC to add as_function, allowing for all PostProcesses to be called as functions
    almost directly.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def as_function(cls, data, *args, **kwargs):
        # Find the parameter's default
        parameters = cls.default_parameters(*args, **kwargs)
        return cls(parameters=parameters).run(data)

    @classmethod
    def default_parameters(cls, *args, **kwargs):
        return get_parameters(cls.__name__).default(*args, **kwargs)


def get_process(process, suffix=""):
    """
    Dynamically import a process class from the available process locations.
    Assumes process filename and class name are the same

    suffix : str
        Name of suffix, generally "" (empty) or "Parameters".
    """
    base_location = "postprocessor.core"
    possible_locations = ("processes", "multisignal")

    found = None
    for possible_location in possible_locations:
        location = (
            f"{base_location}.{possible_location}.{process}.{process}{suffix}"
        )
        found = locate(location)
        if found is not None:
            return found

    raise Exception(
        f"{process} not found in locations {possible_locations} at {base_location}"
    )


def get_parameters(process):
    """
    Dynamically import parameters from the 'processes' folder.
    Assumes parameter is the same name as the file with 'Parameters' added at the end.
    """
    return get_process(process, suffix="Parameters")



from collections import namedtuple

import numpy as np
import pandas as pd
import scipy.linalg as linalg
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC

# TODO: Provide the option of whether to optimise AR order -- see issue #1
# TODO: Provide the functionality of 'smoothing' a time series with AR -- see
# issue #1


class autoregParameters(ParametersABC):
    """Parameters for the 'autoreg' process

    Parameters for the 'autoreg' process.

    Attributes
    ----------
    sampling_period : float
        Sampling period of measurement values, in unit time.

    freq_npoints : int
        Number of points for the frequency axis of the closed-form solution of
        the estimated periodogram.  Defines the resolution for use in, for
        example, plots.

        # Sampling period should be listed in the HDF5 metadata (i.e. the
        # 'time_settings/timeinterval' attribute, unit seconds).  When using
        # this processor in a typical routine, use the sampling period from
        # there rather than relying on this default value of 5 minutes.
    """

    _defaults = {
        "sampling_period": 5,
        "freq_npoints": 100,
    }


class autoreg(PostProcessABC):
    """
    Process to estimate power spectral density (autoregressive model).

    Methods
    -------
    fit_autoreg(timeseries: array_like, ar_order: int)
        Computes linear algebra parameters used to fit an autoregressive model
        to a time series
    optimise_ar_order(timeseries: array_like, ar_order_upper_limit: int)
        Optimise autoregressive model order to fit a time series
    autoreg_periodogram(timeseries: array_like, sampling_period: float,
        freq_npoints: int, ar_order: int)
        Estimates the closed-form solution of the sample power spectrum of a
        timeseries, based on fitting an autoregressive model.
    run(signal: pd.DataFrame)
        Estimates the power spectral densities of a dataframe of time series,
        using autoregressive model
    """

    def __init__(
        self,
        parameters: autoregParameters,
    ):
        super().__init__(parameters)

    # Should this be a static method?
    def fit_autoreg(self, timeseries, ar_order) -> dict:
        """Computes linear algebra parameters to fit an AR model to a timeseries

        Computes linear algebra parameters used to fit an autoregressive model
        of a specified order to a time series.  Ultimately computes the Akaike
        information criterion of the model with the specified order.

        This model-fitting implementation is based on the supplementary
        information of:
        * Jia & Grima (2020). BioRxiv [https://doi.org/10.1101/2020.09.23.309724]

        Parameters
        ----------
        timeseries : array_like
            Time series of measurement values.
        ar_order : int
            Order of the autoregressive model. The order defines the complexity
            of the model; if the order is P, then each time point is modelled as
            a linear combination of P time points preceding it.

        Returns
        -------
        This function returns a dictionary, whose values are defined by these
        keys:

        sample_acfs : 1d array of floats
            Sample autocorrelation function; element indices go from 0 to
            ar_order.

        ar_coeffs : 1d array of floats
            Coefficients for the autoregressive model; defines the linear
            combination that defines the model.

        noise_param : float
            Noise parameter.

        aic : float
            Akaike information criterion of the autoregressive model.

        """
        # Estimates sample autocorrelation function (R).
        # sample_acfs: 1D array of R values
        sample_acfs = np.zeros(ar_order + 1)
        for ii in range(ar_order + 1):
            sample_acfs[ii] = (1 / len(timeseries)) * np.sum(
                [
                    (timeseries[k] - np.mean(timeseries))
                    * (timeseries[k + ii] - np.mean(timeseries))
                    for k in range(len(timeseries) - ii)
                ]
            )

        # Estimates AR coefficients (phi) by solving Yule-Walker equation.
        # ar_coeffs: 1D array of coefficients (i.e. phi values)
        sample_acfs_toeplitz = linalg.toeplitz(sample_acfs[0:ar_order])
        # phi vector goes from 1 to P in Jia & Grima (2020)
        ar_coeffs = linalg.inv(sample_acfs_toeplitz).dot(
            sample_acfs[1 : ar_order + 1]
        )
        # defines a dummy phi_0 as 1.  This is so that the indices I use in
        # get_noise_param are consistent with Jia & Grima (2020)
        ar_coeffs = np.insert(ar_coeffs, 0, 1.0, axis=0)

        # Estimates noise parameter (noise_param)
        noise_param = sample_acfs[0] - np.sum(
            [ar_coeffs[k] * sample_acfs[k] for k in range(1, ar_order + 1)]
        )

        # Calculates AIC (aic)
        aic = np.log(noise_param) + (ar_order) / len(timeseries)

        return {
            "sample_acfs": sample_acfs,
            "ar_coeffs": ar_coeffs,
            "noise_param": noise_param,
            "aic": aic,
        }

    # Should this be a static method?
    def optimise_ar_order(
        self,
        timeseries,
        ar_order_upper_limit,
    ) -> int:
        """Optimise autoregressive model order to fit a time series

        Optimise the autoregressive model order to fit a time series. Model
        selection relies on minimising the Akaike information criterion,
        sweeping over a range of possible orders.

        This implementation is based on the supplementary
        information of:
        * Jia & Grima (2020). BioRxiv [https://doi.org/10.1101/2020.09.23.309724]

        Parameters
        ----------
        timeseries : array_like
            Time series of measurement values.
        ar_order_upper_limit : int
            Upper bound for autoregressive model order; recommended to be the
            square root of the length of the time series.

        Returns
        -------
        int
            Optimal order for an autoregressive model that fits the time series.

        """
        # Bug: artificial dip at order 1 if time series is a smooth sinusoid.
        # Will probably need to fix it so that it checks if the minimum also
        # corresponds to a zero derivative.
        ar_orders = np.arange(1, ar_order_upper_limit)
        aics = np.zeros(len(ar_orders))
        for ii, ar_order in enumerate(ar_orders):
            model = self.fit_autoreg(timeseries, ar_order)
            aics[ii] = model["aic"]
        return ar_orders[np.argmin(aics)]

    def autoreg_periodogram(
        self,
        timeseries,
        sampling_period,
        freq_npoints,
        ar_order,
    ):
        """Estimates the power spectrum of a timeseries, based on an AR model

        Estimates the closed-form solution of the sample power spectrum of a
        time series. This estimation is based on fitting an autoregressive model
        of an optimised order to the time series, and using computed linear
        algebra parameters to compute the closed-form solution.

        This implementation is based on the supplementary
        information of:
        * Jia & Grima (2020). BioRxiv [https://doi.org/10.1101/2020.09.23.309724]

        Parameters
        ---------
        timeseries : array_like
            Time series of measurement values.
        sampling_period : float
            Sampling period of measurement values, in unit time.
        freq_npoints : int
            Number of points for the frequency axis of the closed-form solution of
            the estimated periodogram.  Defines the resolution for use in, for
            example, plots.
        ar_order : int
            Order of the autoregressive model. The order defines the complexity
            of the model; if the order is P, then each time point is modelled as
            a linear combination of P time points preceding it.

        Returns
        -------
        freqs: ndarray
            Array of sample frequencies, unit time-1.

        power: ndarray
            Power spectral density or power spectrum of the time series,
            arbitrary units.

        Examples
        --------
        FIXME: Add docs.

        """
        ar_model = self.fit_autoreg(timeseries, ar_order)
        freqs = np.linspace(0, 1 / (2 * sampling_period), freq_npoints)
        power = np.zeros(len(freqs))
        # Sweeps the frequencies; corresponds to the 'xi' variable in
        # Jia & Grima (2020)
        for ii, freq in enumerate(freqs):
            # multiplied 2pi into the exponential to get the frequency rather
            # than angular frequency
            summation = [
                ar_model["ar_coeffs"][k] * np.exp(-1j * k * (2 * np.pi) * freq)
                for k in range(ar_order + 1)
            ]
            summation[0] = 1  # minus sign error???
            divisor = np.sum(summation)
            power[ii] = (ar_model["noise_param"] / (2 * np.pi)) / np.power(
                np.abs(divisor), 2
            )
        # Normalise by first element of power axis.  This is consistent with
        # the MATLAB code from Ramon Grima.
        power = power / power[0]
        return freqs, power

    def run(self, signal: pd.DataFrame):
        # TODO: Return an additional DataFrame that contains the orders,
        # optimised or not
        """Estimates the power spectra of a dataframe of time series, using AR

        Estimates the power spectral densities of a dataframe of time series.
        This estimation is based on fitting an autoregressive model
        of an optimised order to the time series, and using computed linear
        algebra parameters to compute the closed-form solution.

        This implementation is based on the supplementary
        information of:
        * Jia & Grima (2020). BioRxiv [https://doi.org/10.1101/2020.09.23.309724]

        Parameters
        ----------
        signal: pandas.DataFrame
            Time series, with rows indicating individiual time series (e.g. from
            each cell), and columns indicating time points.

        Returns
        -------
        freqs_df: pandas.DataFrame
            Sample frequencies from each time series, with labels preserved from
            'signal'.

        power_df: pandas.DataFrame
            Power spectrum from each time series, with labels preserved from
            'signal'.

        order_df: pandas.DataFrame
            Optimised order of the autoregressive model fitted to each time
            series, with labels preserved from 'signal'.

        Examples
        --------
        FIXME: Add docs.

        """
        AutoregAxes = namedtuple("AutoregAxes", ["freqs", "power", "order"])
        # Each element in this list is a named tuple: (freqs, power, order)
        axes = []
        # Use enumerate instead?
        for row_index in range(len(signal)):
            timeseries = signal.iloc[row_index, :].dropna().values
            optimal_ar_order = self.optimise_ar_order(
                timeseries, int(3 * np.sqrt(len(timeseries)))
            )
            axes.insert(
                row_index,
                AutoregAxes(
                    *self.autoreg_periodogram(
                        timeseries=timeseries,
                        sampling_period=self.sampling_period,
                        freq_npoints=self.freq_npoints,
                        ar_order=optimal_ar_order,
                    ),
                    optimal_ar_order,
                ),
            )

        freqs_df = pd.DataFrame(
            [element.freqs for element in axes], index=signal.index
        )
        power_df = pd.DataFrame(
            [element.power for element in axes], index=signal.index
        )
        order_df = pd.DataFrame(
            [element.order for element in axes], index=signal.index
        )

        return freqs_df, power_df, order_df



#!/usr/bin/env python3

import numpy as np
import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class crosscorrParameters(ParametersABC):
    """
    Parameters for the 'align' process.

    Attributes
    ----------
    t: int
        Lag, in time points
        FIXME: clarify with Peter.
    """

    _defaults = {
        "lagtime": None,
    }


class crosscorr(PostProcessABC):
    """ """

    def __init__(self, parameters: crosscorrParameters):
        super().__init__(parameters)

    def run(self, trace_dfA: pd.DataFrame, trace_dfB: pd.DataFrame = None):
        """Calculates normalised cross-correlations as a function of time.

        Calculates normalised auto- or cross-correlations as a function of time.
        Normalisation is by the product of the standard deviation for each
        variable calculated across replicates at each time point.
        With zero lag, the normalised correlation should be one.

        Parameters
        ----------
        trace_dfA: dataframe
            An array of signal values, with each row a replicate measurement
            and each column a time point.
        trace_dfB: dataframe (required for cross-correlation only)
            An array of signal values, with each row a replicate measurement
            and each column a time point.

        Returns
        -------
        norm_corr: array or aliby dataframe
            An array of the correlations with each row the result for the
            corresponding replicate and each column a time point
        """

        df = (
            trace_dfA.copy()
            if type(trace_dfA) == pd.core.frame.DataFrame
            else None
        )
        # convert from aliby dataframe to arrays
        trace_A = trace_dfA.to_numpy()
        # number of time points
        n_tps = trace_A.shape[1]
        # number of replicates
        n_replicates = trace_A.shape[0]
        # deviation from mean at each time point
        dmean_A = trace_A - np.nanmean(trace_A, axis=0).reshape((1, n_tps))
        # standard deviation over time for each replicate
        stdA = np.sqrt(
            np.nanmean(dmean_A**2, axis=1).reshape((n_replicates, 1))
        )
        if trace_dfB is not None:
            trace_B = trace_dfB.to_numpy()
            # cross correlation
            dmean_B = trace_B - np.nanmean(trace_B, axis=0).reshape((1, n_tps))
            stdB = np.sqrt(
                np.nanmean(dmean_B**2, axis=1).reshape((n_replicates, 1))
            )
        else:
            # auto correlation
            dmean_B = dmean_A
            stdB = stdA
        # calculate correlation
        corr = np.nan * np.ones(trace_A.shape)
        # lag r runs over time points
        for r in np.arange(0, n_tps):
            prods = [
                dmean_A[:, self.lagtime] * dmean_B[:, self.lagtime + r]
                for self.lagtime in range(n_tps - r)
            ]
            corr[:, r] = np.nansum(prods, axis=0) / (n_tps - r)
        norm_corr = np.array(corr) / stdA / stdB
        # return as a df if trace_dfA is a df else as an array
        return pd.DataFrame(norm_corr, index=df.index, columns=df.columns)



"""Gaussian process fit of a Signal."""
import logging

import gaussianprocessderivatives as gp
import numpy as np
import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


def estimate_gr(volume, dt, noruns, bounds, verbose):
    """
    Parameters
    ----------

    volume : pd.Series
        The volume series of a given cell
    dt : float
        The time interval in hours
    noruns : int
        The number of runs for optimisation
    bounds : dict
        The hyperparameter bounds used for optimisation
    verbose : bool
        If True, prints results

    Returns
    -------
    """
    volume = volume.values
    n = len(volume)
    idx = np.arange(n)
    t = idx * dt
    y = volume[volume > 0]
    x = t[volume > 0]
    idx = idx[volume > 0]
    # Fit the GP
    mg = gp.maternGP(bounds, x, y)
    mg.findhyperparameters(noruns=noruns)
    if verbose:
        mg.results()  # Prints out the hyperparameters
    mg.predict(x, derivs=2)  # Saves the predictions to object
    # Save the predictions to a csv file so they can be reused
    results = dict(
        time=mg.x,
        volume=mg.y,
        fit_time=mg.xnew,
        fit_volume=mg.f,
        growth_rate=mg.df,
        d_growth_rate=mg.ddf,
        volume_var=mg.fvar,
        growth_rate_var=mg.dfvar,
        d_growth_rate_var=mg.ddfvar,
    )
    for name, value in results.items():
        results[name] = np.full((n,), np.nan)
        results[name][idx] = value
    return results


# Give that to a writer: NOTE the writer needs to learn how to write the
# output of a process that returns multiple signals like this one does.


class gpsignalParameters(ParametersABC):
    """
    Parameters
    ----------
        dt : float
            The time step between time points, in minutes
        noruns : int
            The number of times the optimisation is tried
        bounds : dict
            Hyperparameter bounds for the Matern Kernel
        verbose : bool
            Determines whether to print hyperparameter results
    """

    _defaults = dict(
        dt=5,
        noruns=5,
        bounds={0: (-2, 3), 1: (-2, 1), 2: (-4, -1)},
        verbose=False,
    )


class gpsignal(PostProcessABC):
    """Gaussian process fit of a Signal."""

    def __init__(self, parameters: gpsignalParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        results = signal.apply(
            lambda x: estimate_gr(x, **self.parameters.to_dict()),
            result_type="expand",
            axis=1,
        )
        multi_signal = {
            name: pd.DataFrame(
                np.vstack(results[name]),
                index=signal.index,
                columns=signal.columns,
            )
            for name in results.columns
        }
        return multi_signal



import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class dsignalParameters(ParametersABC):
    """
    :window: Number of timepoints to consider for signal.
    """

    _defaults = {"window": 3}


class dsignal(PostProcessABC):
    """
    Calculate the change in a signal depending on a window
    """

    def __init__(self, parameters: dsignalParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        return (
            signal.rolling(window=self.parameters.window, axis=1)
            .mean()
            .diff(axis=1)
        )



from abc import ABC, abstractmethod

# from copy import copy
from itertools import groupby
from typing import List, Tuple, Union

import igraph as ig
import numpy as np
import pandas as pd
from agora.abc import ParametersABC
from agora.io.cells import Cells
from utils_find_1st import cmp_equal, find_1st

from postprocessor.core.abc import PostProcessABC
from postprocessor.core.functions.tracks import max_nonstop_ntps, max_ntps


class pickerParameters(ParametersABC):
    _defaults = {
        "sequence": [
            ["lineage", "intersection", "families"],
            # ["condition", "intersection", "any_present", 0.7],
            # ["condition", "intersection", "growing", 80],
            ["condition", "intersection", "present", 7],
            # ["condition", "intersection", "mb_guess", 3, 0.7],
            # ("lineage", "intersection", "full_families"),
        ],
    }


class picker(PostProcessABC):
    """
    :cells: Cell object passed to the constructor
    :condition: Tuple with condition and associated parameter(s), conditions can be
    "present", "nonstoply_present" or "quantile".
    Determines the thersholds or fractions of signals/signals to use.
    :lineage: str {"mothers", "daughters", "families" (mothers AND daughters), "orphans"}. Mothers/daughters picks cells with those tags, families pick the union of both and orphans the difference between the total and families.
    """

    def __init__(
        self,
        parameters: pickerParameters,
        cells: Cells,
    ):
        super().__init__(parameters=parameters)

        self._cells = cells

    def pick_by_lineage(self, signals, how):
        self.orig_signals = signals

        idx = signals.index

        if how:
            mothers = set(self.mothers)
            daughters = set(self.daughters)
            # daughters, mothers = np.where(mother_bud_mat)

            def search(a, b):
                return np.where(
                    np.in1d(
                        np.ravel_multi_index(
                            np.array(a).T, np.array(a).max(0) + 1
                        ),
                        np.ravel_multi_index(
                            np.array(b).T, np.array(a).max(0) + 1
                        ),
                    )
                )

            if how == "mothers":
                idx = mothers
            elif how == "daughters":
                idx = daughters
            elif how == "daughters_w_mothers":
                present_mothers = idx.intersection(mothers)
                idx = set(
                    [
                        tuple(x)
                        for m in present_mothers
                        for x in np.array(self.daughters)[
                            search(self.mothers, m)
                        ]
                    ]
                )

                print("associated daughters: ", idx)
            elif how == "mothers_w_daughters":
                present_daughters = idx.intersection(daughters)
                idx = set(
                    [
                        tuple(x)
                        for d in present_daughters
                        for x in np.array(self.mothers)[
                            search(self.daughters, d)
                        ]
                    ]
                )
            elif how == "full_families":
                present_mothers = idx.intersection(mothers)
                dwm_idx = set(
                    [
                        tuple(x)
                        for m in present_mothers
                        for x in np.array(self.daughters)[
                            search(np.array(self.mothers), m)
                        ]
                    ]
                )
                present_daughters = idx.intersection(daughters)
                mwd_idx = set(
                    [
                        tuple(x)
                        for d in present_daughters
                        for x in np.array(self.mothers)[
                            search(np.array(self.daughters), d)
                        ]
                    ]
                )
                idx = mwd_idx.union(dwm_idx)
            elif how == "families" or how == "orphans":
                families = mothers.union(daughters)
                if how == "families":
                    idx = families
                elif how == "orphans":
                    idx = idx.diference(families)

            idx = idx.intersection(signals.index)

        return idx

    def pick_by_condition(self, signals, condition, thresh):
        idx = self.switch_case(signals, condition, thresh)
        return idx

    def get_mothers_daughters(self):
        ma = self._cells["mother_assign_dynamic"]
        trap = self._cells["trap"]
        label = self._cells["cell_label"]
        nested_massign = self._cells.mother_assign_from_dynamic(
            ma, label, trap, self._cells.ntraps
        )
        # mother_bud_mat = self.mother_assign_to_mb_matrix(nested_massign)

        if sum([x for y in nested_massign for x in y]):

            mothers, daughters = zip(
                *[
                    ((tid, m), (tid, d))
                    for tid, trapcells in enumerate(nested_massign)
                    for d, m in enumerate(trapcells, 1)
                    if m
                ]
            )
        else:
            mothers, daughters = ([], [])
            print("Warning:Picker: No mother-daughters assigned")

        return mothers, daughters

    def run(self, signals):
        self.orig_signals = signals
        indices = set(signals.index)
        self.mothers, self.daughters = self.get_mothers_daughters()
        for alg, op, *params in self.sequence:
            new_indices = tuple()
            if indices:
                if alg == "lineage":
                    param1 = params[0]
                    new_indices = getattr(self, "pick_by_" + alg)(
                        signals.loc[list(indices)], param1
                    )
                else:
                    param1, *param2 = params
                    new_indices = getattr(self, "pick_by_" + alg)(
                        signals.loc[list(indices)], param1, param2
                    )

            if op == "union":
                # new_indices = new_indices.intersection(set(signals.index))
                new_indices = indices.union(new_indices)

            indices = indices.intersection(new_indices)

        return np.array(list(indices))

    def switch_case(
        self,
        signals: pd.DataFrame,
        condition: str,
        threshold: Union[float, int, list],
    ):
        if len(threshold) == 1:
            threshold = [_as_int(*threshold, signals.shape[1])]
        case_mgr = {
            "any_present": lambda s, thresh: any_present(s, thresh),
            "present": lambda s, thresh: s.notna().sum(axis=1) > thresh,
            "nonstoply_present": lambda s, thresh: s.apply(thresh, axis=1)
            > thresh,
            "growing": lambda s, thresh: s.diff(axis=1).sum(axis=1) > thresh,
            "mb_guess": lambda s, p1, p2: self.mb_guess_wrap(s, p1, p2)
            # "quantile": [np.quantile(signals.values[signals.notna()], threshold)],
        }
        return set(signals.index[case_mgr[condition](signals, *threshold)])

    def mb_guess(self, df, ba, trap, min_budgrowth_t, min_mobud_ratio):
        """
        Parameters
        ----------
        signals : pd.DataFrame
        ba : list of cell_labels that come from bud assignment
        trap : Trap id (used to fetch raw bud)
        min_budgrowth_t: Minimal number of timepoints we lock reassignment after assigning bud
        min_initial_size: Minimal mother-bud ratio when it was first identified
        add_ba: Bool that incorporates bud_assignment data after the normal assignment

        Thinking this problem  as the Movie Scheduling problem (Skiena's the algorithm design manual chapter 1.2),
        we will try to pick the set of filtered cells that grow the fastest and don't overlap within 5 time points
        TODO adjust overlap to minutes using metadata
        """

        # if trap == 21: # Use this to check specific trap problems through a debugger
        #     print("stop")
        ntps = df.notna().sum(axis=1)
        mother_id = df.index[ntps.argmax()]
        nomother = df.drop(mother_id)
        if not len(nomother):
            return []
        nomother = nomother.loc[  # Clean short-lived cells outside our mother cell's timepoints
            nomother.apply(
                lambda x: x.first_valid_index()
                >= df.loc[mother_id].first_valid_index()
                and x.first_valid_index()
                <= df.loc[mother_id].last_valid_index(),
                axis=1,
            )
        ]

        score = -nomother.apply(  # Get slope of candidate daughters
            lambda x: self.get_slope(x.dropna()), axis=1
        )
        start = nomother.apply(pd.Series.first_valid_index, axis=1)

        # clean duplicates
        duplicates = start.duplicated(False)
        if duplicates.any():
            score = self.get_nodup_idx(start, score, duplicates, nomother)
            nomother = nomother.loc[score.index]
            nomother.index = nomother.index.astype("int")
            start = start.loc[score.index]
            start.index = start.index.astype(int)

        d_to_mother = (
            nomother[start] - df.loc[mother_id, start] * min_mobud_ratio
        ).sort_index(axis=1)
        size_filter = d_to_mother[
            d_to_mother.apply(lambda x: x.dropna().iloc[0], axis=1) < 0
        ]
        cols_sorted = (
            size_filter.sort_index(axis=1)
            .apply(pd.Series.first_valid_index, axis=1)
            .sort_values()
        )
        score = score.loc[cols_sorted.index]
        if not len(cols_sorted):
            bud_candidates = pd.DataFrame()
        else:
            # Find the set with the highest number of growing cells and highest avg growth rate for this #
            mivs = self.max_ind_vertex_sets(
                cols_sorted.values, min_budgrowth_t
            )
            best_set = list(
                mivs[np.argmin([sum(score.iloc[list(s)]) for s in mivs])]
            )
            best_indices = cols_sorted.index[best_set]

            start = start.loc[best_indices]
            bud_candidates = cols_sorted.loc[best_indices]
            # bud_candidates = cols_sorted.loc[
            #     [True, *(np.diff(cols_sorted.values) > min_budgrowth_t)]
            # ]

        # Add random-forest bud assignment information here
        new_ba_cells = []
        if (
            ba
        ):  # Use the mother-daughter rf information to prioritise tracks over others
            # TODO add merge application to indices and see if that recovers more cells
            ba = set(ba).intersection(nomother.index)
            ba_df = nomother.loc[ba, :]
            start_ba = ba_df.apply(pd.Series.first_valid_index, axis=1)
            new_ba_cells = list(set(start_ba.index).difference(start.index))

            distances = np.subtract.outer(
                start.values, start_ba.loc[new_ba_cells].values
            )
            todrop, _ = np.where(abs(distances) < min_budgrowth_t)
            bud_candidates = bud_candidates.drop(bud_candidates.index[todrop])

        return [mother_id] + bud_candidates.index.tolist() + new_ba_cells

    @staticmethod
    def max_ind_vertex_sets(values, min_distance):
        """
        Generates an adjacency matrix from multiple points, joining neighbours closer than min_distance
        Then returns the maximal independent vertex sets
        values: list of int values
        min_distance: int minimal distance to cluster
        """
        adj = np.zeros((len(values), len(values))).astype(bool)
        dist = abs(np.subtract.outer(values, values))
        adj[dist <= min_distance] = True

        g = ig.Graph.Adjacency(adj, mode="undirected")
        miv_sets = g.maximal_independent_vertex_sets()
        return miv_sets

    def get_nodup_idx(self, start, score, duplicates, nomother):
        """
        Return the start DataFrame without duplicates

        :start: pd.Series indicating the first valid time point
        :score: pd.Series containing a score to minimise
        :duplicates: Dataframe containing duplicated entries
        :nomother: Dataframe with non-mother cells
        """
        dup_tps = np.unique(start[duplicates])
        idx, tps = zip(
            *[
                (score.loc[nomother.loc[start == tp, tp].index].idxmin(), tp)
                for tp in dup_tps
            ]
        )
        score = score[~duplicates]
        score = pd.concat(
            (score, pd.Series(tps, index=idx, dtype="int", name="cell_label"))
        )
        return score

    def mb_guess_wrap(self, signals, *args):
        if not len(signals):
            return pd.Series([])
        ids = []
        mothers, buds = self.get_mothers_daughters()
        mothers = np.array(mothers)
        buds = np.array(buds)
        ba = []
        # if buds.any():
        #     ba_bytrap = {
        #         i: np.where(buds[:, 0] == i) for i in range(buds[:, 0].max() + 1)
        #     }
        for trap in signals.index.unique(level="trap"):
            # ba = list(
            #     set(mothers[ba_bytrap[trap], 1][0].tolist()).union(
            #         buds[ba_bytrap[trap], 1][0].tolist()
            #     )
            # )
            df = signals.loc[trap]
            selected_ids = self.mb_guess(df, ba, trap, *args)
            ids += [(trap, i) for i in selected_ids]

        idx_srs = pd.Series(False, signals.index).astype(bool)
        idx_srs.loc[ids] = True
        return idx_srs

    @staticmethod
    def get_slope(x):
        return np.polyfit(range(len(x)), x, 1)[0]


def _as_int(threshold: Union[float, int], ntps: int):
    if type(threshold) is float:
        threshold = ntps * threshold
    return threshold


def any_present(signals, threshold):
    """
    Returns a mask for cells, True if there is a cell in that trap that was present for more than :threshold: timepoints.
    """
    any_present = pd.Series(
        np.sum(
            [
                np.isin([x[0] for x in signals.index], i) & v
                for i, v in (signals.notna().sum(axis=1) > threshold)
                .groupby("trap")
                .any()
                .items()
            ],
            axis=0,
        ).astype(bool),
        index=signals.index,
    )
    return any_present



#!/usr/bin/env python3

import matplotlib.pyplot as plt

from postprocessor.routines.plottingabc import BasePlotter


class _SinglePlotter(BasePlotter):
    """Draw a line plot of a single time series."""

    def __init__(
        self,
        trace_timepoints,
        trace_values,
        trace_name,
        unit_scaling,
        trace_color,
        trace_linestyle,
        xlabel,
        plot_title,
    ):
        super().__init__(trace_name, unit_scaling, xlabel, plot_title)
        # Define attributes from arguments
        self.trace_timepoints = trace_timepoints
        self.trace_values = trace_values
        self.trace_color = trace_color
        self.trace_linestyle = trace_linestyle

        # Define some labels
        self.ylabel = "Normalised " + self.trace_name + " fluorescence (AU)"

    def plot(self, ax):
        """Draw the line plot on the provided Axes."""
        super().plot(ax)
        ax.plot(
            self.trace_timepoints * self.unit_scaling,
            self.trace_values,
            color=self.trace_color,
            linestyle=self.trace_linestyle,
            label=self.trace_name + " fluorescence",
        )


def single_plot(
    trace_timepoints,
    trace_values,
    trace_name="flavin",
    unit_scaling=1,
    trace_color="b",
    trace_linestyle="-",
    xlabel="Time (min)",
    plot_title="",
    ax=None,
):
    """Plot time series of trace.

    Parameters
    ----------
    trace_timepoints : array_like
        Time points (as opposed to the actual times in time units).
    trace_values : array_like
        Trace to plot.
    trace_name : string
        Name of trace being plotted, e.g. 'flavin'.
    unit_scaling : int or float
        Unit scaling factor, e.g. 1/60 to convert minutes to hours.
    trace_color : string
        matplotlib colour string, specifies colour of line plot.
    trace_linestyle : string
        matplotlib linestyle argument.
    xlabel : string
        x axis label.
    plot_title : string
        Plot title.
    ax : matplotlib Axes
        Axes in which to draw the plot, otherwise use the currently active Axes.

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the plot.

    Examples
    --------
    FIXME: Add docs.

    """
    plotter = _SinglePlotter(
        trace_timepoints,
        trace_values,
        trace_name,
        unit_scaling,
        trace_color,
        trace_linestyle,
        xlabel,
        plot_title,
    )
    if ax is None:
        ax = plt.gca()
    plotter.plot(ax)
    return ax



#!/usr/bin/env python3

import matplotlib.pyplot as plt
import numpy as np

from postprocessor.routines.plottingabc import BasePlotter


class _MeanPlotter(BasePlotter):
    """Draw mean time series plus standard error."""

    def __init__(
        self,
        trace_df,
        trace_name,
        unit_scaling,
        label,
        mean_color,
        error_color,
        mean_linestyle,
        xlabel,
        ylabel,
        plot_title,
    ):
        super().__init__(trace_name, unit_scaling, xlabel, plot_title)
        # Define attributes from arguments
        self.trace_df = trace_df
        self.label = label
        self.mean_color = mean_color
        self.error_color = error_color
        self.mean_linestyle = mean_linestyle

        # Define some labels
        # self.ylabel = "Normalised " + self.trace_name + " fluorescence (AU)"
        self.ylabel = ylabel

        # Mean and standard error
        self.trace_time = np.array(self.trace_df.columns) * self.unit_scaling
        self.mean_ts = self.trace_df.mean(axis=0)
        self.stderr = self.trace_df.std(axis=0) / np.sqrt(len(self.trace_df))

    def plot(self, ax):
        """Draw lines and shading on provided Axes."""
        super().plot(ax)
        ax.plot(
            self.trace_time,
            self.mean_ts,
            color=self.mean_color,
            alpha=0.75,
            linestyle=self.mean_linestyle,
            label="Mean, " + self.label,
        )
        ax.fill_between(
            self.trace_time,
            self.mean_ts - self.stderr,
            self.mean_ts + self.stderr,
            color=self.error_color,
            alpha=0.5,
            label="Standard error, " + self.label,
        )
        ax.legend(loc="upper right")


def mean_plot(
    trace_df,
    trace_name="flavin",
    unit_scaling=1,
    label="wild type",
    mean_color="b",
    error_color="lightblue",
    mean_linestyle="-",
    xlabel="Time (min)",
    ylabel="Normalised flavin fluorescence (AU)",
    plot_title="",
    ax=None,
):
    """Plot mean time series of a DataFrame, with standard error shading.

    Parameters
    ----------
    trace_df : pandas.DataFrame
        Time series of traces (rows = cells, columns = time points).
    trace_name : string
        Name of trace being plotted, e.g. 'flavin'.
    unit_scaling : int or float
        Unit scaling factor, e.g. 1/60 to convert minutes to hours.
    label : string
        Name of group being plotted, e.g. a strain name.
    mean_color : string
        matplotlib colour string for the mean trace.
    error_color : string
        matplotlib colour string for the standard error shading.
    mean_linestyle : string
        matplotlib linestyle argument for the mean trace.
    xlabel : string
        x axis label.
    ylabel : string
        y axis label.
    plot_title : string
        Plot title.
    ax : matplotlib Axes
        Axes in which to draw the plot, otherwise use the currently active Axes.

    Examples
    --------
    FIXME: Add docs.

    """
    plotter = _MeanPlotter(
        trace_df,
        trace_name,
        unit_scaling,
        label,
        mean_color,
        error_color,
        mean_linestyle,
        xlabel,
        ylabel,
        plot_title,
    )
    if ax is None:
        ax = plt.gca()
    plotter.plot(ax)
    return ax



#!/usr/bin/env python3


import pandas as pd

from agora.abc import ParametersABC
from postprocessor.core.abc import PostProcessABC


class mednormParameters(ParametersABC):
    """:window: Number of timepoints to consider for signal."""

    _defaults = {"window": 3}


class mednorm(PostProcessABC):
    """Normalise a channel by the median."""

    def __init__(self, parameters: mednormParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        return (
            signal.rolling(window=self.parameters.window, axis=1)
            .mean()
            .diff(axis=1)
        )
